local ksr = LibStub("AceAddon-3.0"):NewAddon("KeystoneRunner", "AceBucket-3.0", "AceEvent-3.0")

local MYTHIC_KEYSTONE_ID = 138019
local MIN_REPLY_INTERVAL = 60
local MIN_HINT_INTERVAL = 600
local msgSep_log = "===== Weekly Mythic+ Log ====="
local msgLogEntryID = " [%s] "

local msgHeadnote_new = "Keystone Runner 報告新鑰石："
local msgHeadnote_all = "Keystone Runner 報告所有鑰石："
local msgListEmpty = " · 目前沒有任何鑰石被記錄下來哦。"
local msgDespFullAutoReply = "Keystone Runner 發現對話中提到了「%s」並因此自動發送了鑰石列表。< 1分鐘之內不會再次觸發 >"
local msgHintSemiAutoReply = "Keystone Runner 發現對話中提到了「%s」，使用 /ksr r 即可在對應的密語或小隊頻道報告所有鑰石。< 10分鐘之內不會再次提示 >"
local msgWeeklyCleanup = "Keystone Runner 新的一週開始了，上週的數據已經清除。"
local msgWeeklyBest = "本周最佳"

local msgLogSum = "找到了 %s 條M+挑戰日誌 <<"
local msgLogEmpty = " · 目前沒有任何M+挑戰被記錄下來哦。"
local msgLogEntryBody = "%s %s(+%s) { %s } 耗時：%s 鑰石升級：+%s"

local msgUsage = "Keystone Runner 記錄並報告所有角色持有的鑰石。\r* 使用 /ksr p、/ksr g 報告到小隊或公會頻道。\r* 使用 /ksr help 查看更多幫助信息。"
-- \r* 檢測到和「鑰石」有關的密語時，提示並輔助密語回覆鑰石列表。
-- \r* 使用 /ksr w [玩家名字-伺服器] 密語報告給特定玩家。

local kwKeywords = "keystone|鑰石|钥石|key|鑰匙|钥匙|m%+|大秘|大米|保底|低保"
local kwFilters = "Keystone Runner| · "

-- Utils

function ksr:log(msg)
	print(msg)
end

function ksr:isLeft(str, left)
	return string.sub(str, 1, string.len(left)) == left
end

function ksr:nameWithRealm(player, realm)
	if player == nil then
		return "[N/A]"
	end
	
	if (realm == nil) or (realm == "") then
		return format("%s-%s", player, GetRealmName())
	else
		return format("%s-%s", player, realm)
	end
end

function ksr:printUsage()
	local usage = { strsplit("\r", msgUsage) }
	for _, v in pairs(usage) do
		self:log(v)
	end
end

-- Events

function ksr:onBagUpdate()
	self:checkNewKeyStone()
end

function ksr:onChallengeModeCompleted()
	self:checkNewKeyStone()

	local _, level, elapsedMS, _, keystoneUpgradeLevels = C_ChallengeMode.GetCompletionInfo()
	self:updateWeeklyBest(level)
	
	-- log m+ data
	local dateTime = date("%Y/%m/%d %H:%M")
	local mapName = GetInstanceInfo()
	local partyMember = format("%s %s %s %s %s", self.name, self:nameWithRealm(UnitName("party1")), self:nameWithRealm(UnitName("party2")), self:nameWithRealm(UnitName("party3")), self:nameWithRealm(UnitName("party4")))
	local elapsedTime = SecondsToTime(elapsedMS / 1000)
	
	local newLogEntry = {name = self.name, dateTime = dateTime, mapName = mapName, level = level, partyMember = partyMember, elapsedTime = elapsedTime, keystoneUpgradeLevels = keystoneUpgradeLevels}
	table.insert(self.MPlusLog, newLogEntry)
end

function ksr:onChatMsg(event, ...)
	local argv = {...}
	local ID
	local channel
	local msg = argv[1]	
	
	if event == "CHAT_MSG_BN_WHISPER" then
		ID = argv[13]
		channel = "BN_WHISPER"
	elseif event == "CHAT_MSG_WHISPER" then
		ID = argv[2]
		channel = "WHISPER"
	elseif event == "CHAT_MSG_PARTY" or event == "CHAT_MSG_PARTY_LEADER" then
		if argv[2] == self.name then
			return
		end
		ID = "_PARTY_"
		channel = "PARTY"
	end

	if self:checkReplyInterval(ID) and self:checkFilters(msg) then
		local hasKeyword, keyword = self:checkKeywords(msg:lower())
		if not hasKeyword then
			return
		end
		
		if self.fullAutoReply then -- or isAFK ?
			self:announceAllKeystones(channel, ID, true, keyword)
			self:updateReplyInterval(ID)
		else
			if GetTime() - self.MRT_Time >= MIN_HINT_INTERVAL then
				self:log(format(msgHintSemiAutoReply, keyword))
				self.MRT_Time = GetTime()
			end
			self.MRT_Channel = channel
			self.MRT_ID = ID
		end
	end
end

-- Functions

function ksr:checkNewKeyStone()
	local keystone, changed = self:updateKeystone()
	if changed then
		self:announceKeystone(keystone)
	end
end

function ksr:checkReplyInterval(ID)
	if (self.timeReply[ID] == nil) or (GetTime() - self.timeReply[ID] >= MIN_REPLY_INTERVAL) then	
		return true
	else
		--self:log(msgReplyCooldown)
		return false
	end
end

function ksr:checkFilters(msg)
	for _, v in pairs(self.Filters) do
		if self:isLeft(msg, v) then
			return false
		end
	end
	
	return true
end

function ksr:checkKeywords(msg)
	for _, v in pairs(self.Keywords) do
		if strfind(msg, v) then
			return true, v
		end
	end
	
	return false, nil
end

function ksr:updateReplyInterval(ID)
	self.timeReply[ID] = GetTime()
end

function ksr:updateWeeklyBest(level)
	if (self.WeeklyBest[self.name] == nil) or (self.WeeklyBest[self.name] < level) then
		self.WeeklyBest[self.name] = level
	end
end

function ksr:textOfKeystone(keystone, plainText)
	local msg
	local classColor = RAID_CLASS_COLORS[keystone.classE]
	if classColor and not plainText then
		msg = format("|c%s%s|r", classColor.colorStr, keystone.name)
	else
		msg = keystone.name
	end
	return msg..format("(%s) %s(+%s)", keystone.class, keystone.link, keystone.keystoneLevel)
end

function ksr:textOfWeeklyBest(keystone)
	local weeklyBest = self.WeeklyBest[keystone.name]
	if (weeklyBest == nil) or (weeklyBest == 0) then
		return "N/A"
	else
		return format("+%s", weeklyBest)
	end
end

function ksr:addChatMessage(msg, channel, ID)
	if (channel == "PARTY" and IsInGroup()) or (channel == "GUILD" and GetGuildInfo("player")) or (channel == "SAY") then
		SendChatMessage(msg, channel)
		return true
	elseif (channel == "WHISPER") and (ID ~= nil) then
		SendChatMessage(msg, channel, nil, ID)
		return true
	elseif (channel == "BN_WHISPER") and (ID ~= nil) then
		BNSendWhisper(ID, msg)
		return true
	else
		self:log(msg)
		return false
	end
end

function ksr:announceKeystone(keystone)
	-- announce keystone to party members
	if keystone then
		local msgKS = msgHeadnote_new..self:textOfKeystone(keystone, true)
		return self:addChatMessage(msgKS, "PARTY", nil)
	else
		return false
	end
end

function ksr:announceAllKeystones(channel, ID, autoReply, keyword)
	local retVal
	local plainText = true
	local sorted = {}
	
	if channel == "DEFCHAT" then
		plainText = false
	end
	
	retVal = self:addChatMessage(msgHeadnote_all, channel, ID)
	
	if next(self.Keystones) ~= nil then
		for _, ks in pairs(self.Keystones) do
			table.insert(sorted, ks)
		end

		-- order by keystoneLevel desc
		table.sort(sorted, function(a, b) return a.keystoneLevel > b.keystoneLevel end)
	
		for i = 1, #sorted do
			local msg = " · "..self:textOfKeystone(sorted[i], plainText)..format(" [%s %s]", msgWeeklyBest, self:textOfWeeklyBest(sorted[i]))
			retVal = self:addChatMessage(msg, channel, ID) and retVal
		end
	else
		retVal = self:addChatMessage(msgListEmpty, channel, ID) and retVal
	end
	
	if autoReply then
		retVal = self:addChatMessage(format(msgDespFullAutoReply, keyword), channel, ID) and retVal
	end
	
	if not retVal then
		self:printUsage()
	end
end

function ksr:weeklyCleanUp()
	C_ChallengeMode.RequestRewards()

	if C_ChallengeMode.IsWeeklyRewardAvailable() and (next(self.Keystones) ~= nil) then
		self:resetData()
		return true
	else
		return false
	end
end

function ksr:updateKeystone()
	-- do not execute update during ChallengeMode, otherwise will get an invalid msg "active keystone (level)-1"
	if C_ChallengeMode.IsChallengeModeActive() then
		return nil, nil
	end
	
	if self:weeklyCleanUp() then
		self:log(msgWeeklyCleanup)
		return nil, nil
	end

	local class, classE = UnitClass("player")

	for bag = 0, NUM_BAG_SLOTS do
		local numSlots = GetContainerNumSlots(bag)
		for slot = 1, numSlots do
			if GetContainerItemID(bag, slot) == MYTHIC_KEYSTONE_ID then
				local orgLink = GetContainerItemLink(bag, slot)
				local parts = { strsplit(":", orgLink) }

				local dungeonID = tonumber(parts[2])
				local keystoneLevel = tonumber(parts[3])

				local newKeystone = {name = self.name, class = class, classE = classE, link = orgLink, dungeonID = dungeonID, keystoneLevel = keystoneLevel}
				local oldKeystone = self.Keystones[self.name]
				local changed = ((oldKeystone == nil) or (oldKeystone.keystoneLevel ~= newKeystone.keystoneLevel) or (oldKeystone.dungeonID ~= newKeystone.dungeonID))

				if changed then
					self.Keystones[self.name] = newKeystone
				end
				return self.Keystones[self.name], changed
			end
		end
	end
	-- no keystone
	return nil, nil
end

function ksr:resetData()
	wipe(self.Keystones)
	wipe(self.WeeklyBest)
	wipe(self.MPlusLog)
end

function ksr:viewMPlusLog(filters)
	self:log(msgSep_log)
	
	if next(self.MPlusLog) == nil then
		self:log(msgLogEmpty)
	end
	
	local logCount = 0
	for i = 1, #self.MPlusLog do
		local validEntry = true
		local item = self.MPlusLog[i]
		local msgBody = format(msgLogEntryBody, item.dateTime, item.mapName, item.level, item.partyMember, item.elapsedTime, item.keystoneUpgradeLevels)
		
		logCount = logCount + 1
		
		if #filters >= 2 then
			for k = 2, #filters do
				if not strfind(msgBody, filters[k]) then
					validEntry = false
					logCount = logCount - 1
					break
				end
			end
		end
		
		if validEntry then
			self:log(format(msgLogEntryID, logCount)..msgBody)
		end
	end
	
	self:log(format(msgLogSum, logCount))
end

function ksr:wipelog()
	wipe(self.MPlusLog)
end

function ksr:manualReply()
	if self.MRT_Channel ~= "" and self.MRT_ID ~= "" then
		self:announceAllKeystones(self.MRT_Channel, self.MRT_ID, false, nil)
	end
end

function ksr:slashCmd(cmd)
	local cmdp = { strsplit(" ", cmd) }
	-- cant send BN whisper by command
	if cmd == "p" or cmd == "party" then
		self:announceAllKeystones("PARTY", nil, false)
	elseif cmd == "g" or cmd == "guild" then
		self:announceAllKeystones("GUILD", nil, false)
	elseif cmd == "s" or cmd == "say" then
		self:announceAllKeystones("SAY", nil, false)
	elseif cmdp[1] == "w" or cmdp[1] == "whisper" then
		if #cmdp == 2 then
			self:announceAllKeystones("WHISPER", cmdp[2], false)
		end
	elseif cmd == "r" or cmd == "reply" then
		self:manualReply()
	-- advanced commands
	elseif cmd == "clear" then
		self:resetData()
	elseif cmdp[1] == "log" then
		self:viewMPlusLog(cmdp)
	elseif cmd == "wipelog" then
		self:wipelog()
	else
		self:announceAllKeystones("DEFCHAT", nil, false)
	end
end

-- init

function ksr:initKeywords()
	self.Keywords = { strsplit("|", kwKeywords) }
	self.Filters = { strsplit("|", kwFilters) }
end

function ksr:initWeeklyBest()
	C_ChallengeMode.RequestMapInfo()
	
	for _, mapID in pairs(C_ChallengeMode.GetMapTable()) do
		local _, _, weeklyBestLevel = C_ChallengeMode.GetMapPlayerStats(mapID)
		if weeklyBestLevel then
			self:updateWeeklyBest(weeklyBestLevel)
		end
	end
end

function ksr:registerEvent(keystone)
	local interval = 3
	
	if keystone ~= nil then
		-- set bag scanning interval to 1 min in case of unexpected events
		interval = 60
	end
	self:RegisterBucketEvent("BAG_UPDATE", interval, "onBagUpdate")
	self:RegisterBucketEvent("CHALLENGE_MODE_COMPLETED", 3, "onChallengeModeCompleted")

	self.chatFrame = CreateFrame("Frame")
	self.chatFrame:RegisterEvent("CHAT_MSG_WHISPER")
	self.chatFrame:RegisterEvent("CHAT_MSG_BN_WHISPER")
	self.chatFrame:RegisterEvent("CHAT_MSG_PARTY")
	self.chatFrame:RegisterEvent("CHAT_MSG_PARTY_LEADER")
	self.chatFrame:SetScript("OnEvent", function(obj, event, ...) self:onChatMsg(event, ...) end)
end

function ksr:OnInitialize()
	self.fullAutoReply = false
	self.MRT_Time = 0
	self.MRT_Channel = ""
	self.MRT_ID = ""

	self.chatFrame = nil
	self.Keywords = {}
	self.Filters = {}
	self.timeReply = {}
	self.name = self:nameWithRealm(UnitName("player"))
	self:initKeywords()
	
	-- init db
	self.Keystones = {}
	self.WeeklyBest = {}
	self.MPlusLog = {}
	local dbDefaults = {
		keystones = {},
		weeklybest = {},
		mpluslog = {},
	}
	self.db = LibStub("AceDB-3.0"):New("KeystoneRunnerDB", {faction = dbDefaults}, true).faction
	self.Keystones = self.db.keystones
	self.WeeklyBest = self.db.weeklybest
	self.MPlusLog = self.db.mpluslog
	
	-- init keystone, events and player data
	-- wait for player login
	self:RegisterBucketEvent("PLAYER_LOGIN", 3, function()
		local keystone, _ = self:updateKeystone()
		self:registerEvent(keystone)
		self:initWeeklyBest()
		self:printUsage()
	end)

	-- misc
	SLASH_KEYSTONERUNNER1, SLASH_KEYSTONERUNNER2 = "/keystonerunner", "/ksr"
	SlashCmdList["KEYSTONERUNNER"] = function(cmd) self:slashCmd(string.lower(cmd)) end	
end
